@@ -1,4 +1,4 @@
-# Copyright 2023 The Flax Authors.
+# Copyright 2024 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,13 +14,13 @@
 
 """Utilities for working with jit and partitioned models.
 
-This module introduces `axis_rules`, `logical_to_mesh_axes`,
-`logical_to_mesh`, `with_logical_constraint` for appyling jit sharding
+This module introduces ``axis_rules``, ``logical_to_mesh_axes``,
+``logical_to_mesh``, ``with_logical_constraint`` for appyling jit sharding
 constraints in terms of "logical named axes" rather than jit's default mesh
 axes.
 
-Additionally the `LogicallyPartitioned` metadata wrapper is defined as
-well as the initializer function wrapper `with_logical_partitioning` for
+Additionally the ``LogicallyPartitioned`` metadata wrapper is defined as
+well as the initializer function wrapper ``with_logical_partitioning``for
 introducing logical axis metadata into a model's variables.
 """
 
@@ -35,17 +35,17 @@ from typing import Any, Callable, List, Optional, Sequence, Tuple, Union
 import jax
 from jax import lax
 from jax.experimental import maps
+
 from flax import struct
-
 from flax.core import meta
-
-# Real types and dummy aliases for documentation
-LogicalRules = Sequence[Tuple[str, Union[str, Tuple[str], None]]]
-Array = Any  # pylint: disable=invalid-name
-ArrayPytree = Any  # pylint: disable=invalid-name
-LogicalPartitionSpec = Any  # pylint: disable=invalid-name
-LogicalPartitionSpecPytree = Any  # pylint: disable=invalid-name
-PartitionSpecPytree = Any  # pylint: disable=invalid-name
+from flax.typing import (
+  Array,
+  LogicalNames,
+  LogicalRules,
+  ArrayPytree,  # pylint: disable=invalid-name
+  LogicalPartitionSpec,  # pylint: disable=unused-import
+  LogicalPartitionSpecPytree,  # pylint: disable=invalid-name
+  )
 
 
 # Dynamic Axis Mapping Context
@@ -107,8 +107,8 @@ def _mesh_assignment_free(new_assignment, existing_assignments):
 
 
 def _logical_to_mesh_axes(
-    array_dim_names: Optional[Sequence[Optional[str]]],
-    rules: Optional[LogicalRules] = None,
+  array_dim_names: Optional[Sequence[Optional[str]]],
+  rules: Optional[LogicalRules] = None,
 ) -> Optional[List[Union[_UnassignedAxis, None, str, Tuple[str]]]]:
   """Same as logical_to_mesh_axes, but doesn't fill in _unassigned_axis."""
   if array_dim_names is None:
@@ -117,11 +117,11 @@ def _logical_to_mesh_axes(
     rules = _axis_rules.rules
   axis_name_counts = collections.Counter(array_dim_names)
   dups = tuple(
-      k for k, v in axis_name_counts.items() if v > 1 and k is not None
+    k for k, v in axis_name_counts.items() if v > 1 and k is not None
   )
   if dups:
     raise ValueError(
-        f'Unsupported: Dimensions {dups} occur more than once in array names.'
+      f'Unsupported: Dimensions {dups} occur more than once in array names.'
     )
   if not isinstance(rules, (tuple, list)):
     raise ValueError('Unknown axis rule specification type.')
@@ -132,21 +132,21 @@ def _logical_to_mesh_axes(
     if rule_model_name in array_dim_names:
       pos = array_dim_names.index(rule_model_name)
       if (
-          _mesh_assignment_free(rule_mesh_names, result)
-          and result[pos] == _unassigned_axis
+        _mesh_assignment_free(rule_mesh_names, result)
+        and result[pos] == _unassigned_axis
       ):
         result[pos] = rule_mesh_names
   return result
 
 
 def logical_to_mesh_axes(
-    array_dim_names: Optional[Sequence[Optional[str]]],
-    rules: Optional[LogicalRules] = None,
+  array_dim_names: Optional[Sequence[Optional[str]]],
+  rules: Optional[LogicalRules] = None,
 ) -> Optional[jax.sharding.PartitionSpec]:
   """Compute layout for an array.
 
   The rules are in order of precedence, and consist of pairs:
-  (ArrayDimensionName, MeshDimensionName), meaning that the given array
+  ``(ArrayDimensionName, MeshDimensionName)``, meaning that the given array
   dimension (if present and unused) should be sharded across the given
   mesh dimension (if present and unused).
 
@@ -155,22 +155,25 @@ def logical_to_mesh_axes(
   mesh-dimension, meaning that this dimension of the array is sharded across
   this dimension of the mesh.
 
-  For example, given an array with
+  For example, given an array with::
+
     array_dim_names = ('batch', 'length', 'heads', 'features')
-  and the layout rules are:
+
+  and the layout rules are::
+
     rules = (('batch', 'X'),
              ('features', 'X'),
              ('heads', 'Y'),
              ('batch', 'Z'))
 
-  then this function will return
+  then this function will return::
 
     PartitionSpec('X', None, 'Y', None)
 
   Args:
     array_dim_names: Tuple of array dimension names or None.
     rules: Optional logical to mesh rules override.  Defaults to using the
-      rules defined in the dynamic context set from the `axis_rules` function.
+      rules defined in the dynamic context set from the ``axis_rules`` function.
 
   Returns:
     PartitionSpec for the parameter.
@@ -186,22 +189,22 @@ def logical_to_mesh_axes(
 def logical_to_mesh(tree: Any, rules: Optional[LogicalRules] = None) -> Any:
   """Applies logical_to_mesh_axes to pytrees of logical PartitionSpecs."""
   return jax.tree_map(
-      lambda x: logical_to_mesh_axes(x, rules),
-      tree,
-      is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
+    lambda x: logical_to_mesh_axes(x, rules),
+    tree,
+    is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
   )
 
 
 def logical_to_mesh_sharding(
-    tree: Any,
-    mesh: jax.sharding.Mesh,
-    rules: Optional[LogicalRules] = None,
+  tree: Any,
+  mesh: jax.sharding.Mesh,
+  rules: Optional[LogicalRules] = None,
 ) -> Any:
   """Convert pytrees of logical PartitionSpecs to shardings."""
   return jax.tree_map(
-      lambda x: jax.sharding.NamedSharding(mesh, x),
-      logical_to_mesh(tree, rules),
-      is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
+    lambda x: jax.sharding.NamedSharding(mesh, x),
+    logical_to_mesh(tree, rules),
+    is_leaf=lambda x: isinstance(x, jax.sharding.PartitionSpec),
   )
 
 
@@ -220,13 +223,13 @@ class RulesFallback(enum.Enum):
 
 
 def _with_sharding_constraint(
-    x: Array,
-    axis_resources: Optional[jax.sharding.PartitionSpec],
-    mesh: Optional[jax.sharding.Mesh] = None,
+  x: Array,
+  axis_resources: Optional[jax.sharding.PartitionSpec],
+  mesh: Optional[jax.sharding.Mesh] = None,
 ):
   """Wrapper for lax.with_sharding_constraint, no-op on cpu or outside jit."""
   if jax.devices()[0].platform == 'cpu' or (
-      not _global_mesh_defined() and mesh is None
+    not _global_mesh_defined() and mesh is None
   ):
     return x
   else:
@@ -237,11 +240,11 @@ def _with_sharding_constraint(
 
 
 def _with_sharding_constraint_one_fallback(
-    axis_resources: LogicalPartitionSpec,
-    x: Array,
-    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
-    rules: Optional[LogicalRules] = None,
-    mesh: Optional[jax.sharding.Mesh] = None,
+  axis_resources: LogicalPartitionSpec,
+  x: Array,
+  fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
+  rules: Optional[LogicalRules] = None,
+  mesh: Optional[jax.sharding.Mesh] = None,
 ):
   """Either imposes a sharding constraint or applies fallback."""
   mesh_axes = _logical_to_mesh_axes(axis_resources, rules)
@@ -257,22 +260,22 @@ def _with_sharding_constraint_one_fallback(
       else:
         return x
   return _with_sharding_constraint(
-      x, jax.sharding.PartitionSpec(*mesh_axes), mesh=mesh
+    x, jax.sharding.PartitionSpec(*mesh_axes), mesh=mesh
   )
 
 
 def _is_logical_spec(x):
   return x is None or (
-      isinstance(x, tuple) and all(isinstance(e, str) or e is None for e in x)
+    isinstance(x, tuple) and all(isinstance(e, str) or e is None for e in x)
   )
 
 
 def with_logical_constraint(
-    x: ArrayPytree,
-    logical_axis_resources: LogicalPartitionSpecPytree,
-    rules: Optional[LogicalRules] = None,
-    mesh: Optional[jax.sharding.Mesh] = None,
-    fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
+  x: ArrayPytree,
+  logical_axis_resources: LogicalPartitionSpecPytree,
+  rules: Optional[LogicalRules] = None,
+  mesh: Optional[jax.sharding.Mesh] = None,
+  fallback: RulesFallback = RulesFallback.AXIS_IS_UNSHARDED,
 ):
   """Version of jit's with_sharding_constraint that uses logical axis names."""
   # If no axis binding is set, this is a no-op.
@@ -282,15 +285,15 @@ def with_logical_constraint(
     return x
   # Translate logical names to mesh assignments.
   return jax.tree_util.tree_map(
-      functools.partial(
-          _with_sharding_constraint_one_fallback,
-          fallback=fallback,
-          rules=rules,
-          mesh=mesh,
-      ),
-      logical_axis_resources,
-      x,
-      is_leaf=_is_logical_spec,
+    functools.partial(
+      _with_sharding_constraint_one_fallback,
+      fallback=fallback,
+      rules=rules,
+      mesh=mesh,
+    ),
+    logical_axis_resources,
+    x,
+    is_leaf=_is_logical_spec,
   )
 
 
@@ -305,28 +308,29 @@ class LogicallyPartitioned(meta.Partitioned):
     """Returns the wrapped value with the partitioning constraint applied."""
     if apply_constraint and (_global_mesh_defined() or self.mesh is not None):
       return with_logical_constraint(
-          self.value,
-          self.get_partition_spec(),
-          rules=self.rules,
-          mesh=self.mesh,
+        self.value,
+        self.get_partition_spec(),
+        rules=self.rules,
+        mesh=self.mesh,
       )
     else:
       return self.value
 
 
 def with_logical_partitioning(
-    fn: Callable[..., Any],
-    names: meta.LogicalNames,
-    mesh: Optional[jax.sharding.Mesh] = None,
-    rules: Optional[LogicalRules] = None,
+  fn: Callable[..., Any],
+  names: LogicalNames,
+  mesh: Optional[jax.sharding.Mesh] = None,
+  rules: Optional[LogicalRules] = None,
 ) -> Callable[..., LogicallyPartitioned]:
   """Wraps a function's return value with LogicallyPartitioned.
 
   Example::
 
-    kernel_init = with_logical_partitioning(
-        nn.initializers.lecun_normal, (None, "data"))
-    partitioned_dense = nn.Dense(features, kernel_init=kernel_init)
+    >>> import flax.linen as nn
+    >>> kernel_init = nn.with_logical_partitioning(
+    ...     nn.initializers.lecun_normal(), (None, "data"))
+    >>> partitioned_dense = nn.Dense(features=3, kernel_init=kernel_init)
 
   Args:
     fn: The function to be wrapped. Typically this is an initializer.
@@ -343,7 +347,7 @@ def with_logical_partitioning(
   @functools.wraps(fn)
   def wrapper(*args, **kwargs):
     return LogicallyPartitioned(
-        fn(*args, **kwargs), names, mesh=mesh, rules=rules
+      fn(*args, **kwargs), names, mesh=mesh, rules=rules
     )  # pytype: disable=wrong-keyword-args
 
   return wrapper
